import math
import unittest

import numpy as np
import torch

from yamlu.bb import iou_matrix, bbs_distances, iou_vector, pts_boxes_distance


class TestBoundingBoxes(unittest.TestCase):
    def test_iou_matrix_no_overlap(self):
        bbs1 = np.array([
            [10, 10, 20, 20],
            [10, 10, 20, 20],
        ])
        bbs2 = np.array([
            [50, 50, 100, 100],
            [50, 50, 100, 100]
        ])
        ious = iou_matrix(bbs1, bbs2)
        assert ious.shape == (2, 2)
        assert np.all(ious == 0.)

    def test_iou_matrix(self):
        bbs1 = np.array([
            [10, 10, 20, 20],
            [10, 10, 30, 30]
        ])
        bbs2 = np.array([
            [10, 10, 20, 20],
            [0, 0, 20, 20],
            [50, 50, 100, 100]
        ])
        ious = iou_matrix(bbs1, bbs2)
        assert ious.shape == (2, 3)
        assert ious[0, 0] == 1.
        assert ious[0, 1] == .25
        assert ious[0, 2] == 0.
        assert ious[1, 0] == .25
        assert ious[1, 1] == 1 / 7
        assert ious[1, 2] == 0

    def test_iou_vector(self):
        bbs1 = np.array([
            [10, 10, 20, 20],
            [0, 0, 40, 40]
        ])
        bbs2 = np.array([
            [10, 10, 20, 20],
            [0, 0, 20, 20],
        ])
        ious = iou_vector(bbs1, bbs2)
        assert ious.shape == (2,)
        assert ious[0] == 1.
        assert ious[1] == .25

    def test_bbs_distances(self):
        bbs1 = np.array([
            [10, 10, 20, 20],
            [10, 10, 30, 30]
        ])
        bbs2 = np.array([
            [10, 10, 20, 20],
            [0, 0, 20, 20],
            [50, 50, 100, 100]
        ])
        distances = bbs_distances(bbs1, bbs2)
        assert distances.shape == (2, 3)
        assert distances[0, 0] == 0
        assert distances[0, 1] == 0
        assert distances[0, 2] == math.sqrt(30 ** 2 + 30 ** 2)
        assert distances[1, 0] == 0
        assert distances[1, 1] == 0
        assert distances[1, 2] == math.sqrt(20 ** 2 + 20 ** 2)

    def test_bbs_distances_intersect(self):
        bbs1 = np.array([
            [10, 0, 20, 50]
        ])
        bbs2 = np.array([
            [0, 10, 14, 30]
        ])
        distances = bbs_distances(bbs1, bbs2)
        assert distances.shape == (1, 1)
        assert distances[0, 0] == 0

    def test_pts_boxes_distance(self):
        pts = torch.tensor([[0, 0], [15, 15]], dtype=torch.int32)

        boxes_ltrb = torch.tensor([
            [10, 10, 20, 20],
            [0, 0, 10, 10]
        ], dtype=torch.float32)

        ds = pts_boxes_distance(pts, boxes_ltrb, zero_dist_pt_within_box=True)
        assert ds[0, 0] == math.sqrt(200)
        assert ds[1, 0] == 0
        print(ds[:, 0])

        ds = pts_boxes_distance(pts, boxes_ltrb, zero_dist_pt_within_box=False)
        assert ds[1, 0] == 5
        print(ds[:, 0])


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from yamlu.img import Annotation, BoundingBox


class TestAnnotation(unittest.TestCase):
    def test_annotation_repr(self):
        a = Annotation("test", BoundingBox(0, 5, 10, 10), text="foo")
        print(a)
        assert str(a) == "Annotation(category='test', bb=BoundingBox(t=0.00,l=5.00,b=10.00,r=10.00), text=foo)"


class TestBoundingBox(unittest.TestCase):
    def test_intersection(self):
        b1 = BoundingBox(0, 0, 10, 10)
        b2 = BoundingBox(0, 0, 5, 5)
        assert b1.intersection(b2) == b2
        assert b1.intersection(b1) == b1

        b3 = BoundingBox(10, 10, 15, 15)
        assert b1.intersection(b3).area == 0.

    def test_from_points(self):
        pts = np.array([
            [0, 2],
            [1, 0]
        ])
        assert BoundingBox.from_points(pts) == BoundingBox(0, 0, 3, 2)

    def test_iou(self):
        b1 = BoundingBox(0, 0, 10, 10)
        b2 = BoundingBox(0, 0, 5, 5)
        assert b1.iou(b2) == 0.25
        assert b1.iou(b1) == 1.

        b3 = BoundingBox(10, 10, 15, 15)
        assert b1.iou(b3) == 0.

        b4 = BoundingBox(5, 5, 15, 15)
        assert b1.iou(b4) == 1 / 7



import unittest

import numpy as np
import torch

from yamlu.regression import regression_metrics


class TestRegressionMetrics(unittest.TestCase):

    def test_torch_perfect_predictions(self):
        y_true = torch.tensor([[0, 5], [1, 2], [1, 4]])
        y_pred = torch.tensor([[0, 5], [1, 2], [1, 4]])
        target_names = ["height", "width"]
        metrics = regression_metrics(y_true, y_pred, target_names=target_names)
        for m in ["mse", "mae"]:
            assert np.all(metrics[m] == 0)



import datetime as dt
import inspect
import logging
import sys
from functools import wraps, partial


# copied from scikit-lego (to avoid the dependencies of scikit-lego)
# https://github.com/koaning/scikit-lego/blob/master/sklego/pandas_utils.py#L14
def log_step(func=None, *, level=logging.INFO):
    """
    Decorates a function that transforms a pandas dataframe to add automated logging statements
    :Example:
    >>> @log_step
    ... def remove_outliers(df, min_obs=5):
    ...     pass
    >>> @log_step(level=logging.INFO)
    ... def remove_outliers(df, min_obs=5):
    ...     pass
    """
    if func is None:
        return partial(log_step, level=level)

    @wraps(func)
    def wrapper(*args, **kwargs):
        logger = logging.getLogger(sys.modules[func.__module__].__name__)

        tic = dt.datetime.now()
        result = func(*args, **kwargs)
        time_taken = str(dt.datetime.now() - tic)
        func_args = inspect.signature(func).bind(*args, **kwargs).arguments
        func_args_str = "".join(
            ", {} = {!r}".format(*item) for item in list(func_args.items())[1:]
        )
        logger.log(
            level,
            f"[{func.__name__}(df{func_args_str})] "
            f"n_obs={result.shape[0]} n_col={result.shape[1]} time={time_taken}",
        )

        return result

    return wrapper



import itertools
import warnings
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn import metrics
from sklearn.preprocessing import LabelEncoder


def cl_report(y_true, y_pred, classes: List[str], digits=3, only_avg=False) -> pd.DataFrame:
    # micro AP == micro AR == accuracy in a multi-class setting (https://datascience.stackexchange.com/a/29054)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        class_rep = metrics.classification_report(y_true, y_pred, labels=list(range(len(classes))),
                                                  target_names=classes, digits=digits, output_dict=True)
    cr_df = pd.DataFrame(class_rep).T
    for col in ['f1-score', 'precision', 'recall']:
        cr_df[col] = cr_df[col].round(decimals=digits)
    if only_avg:
        cr_df = cr_df[cr_df.index.str.contains("avg")]
    return cr_df


def plot_roc_auc(y_true, y_score):
    fpr, tpr, thresholds = metrics.roc_curve(y_true, y_score, pos_label=None, sample_weight=None,
                                             drop_intermediate=True)
    roc_auc = metrics.roc_auc_score(y_true, y_score)

    plt.figure()
    lw = 2
    plt.plot(fpr, tpr, color='darkorange', lw=lw, label='ROC curve (area = %0.2f)' % roc_auc)
    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver operating characteristic example')
    plt.legend(loc="lower right")
    plt.show()


def plot_cm(y_true, y_pred, classes, figsize=(6, 6)):
    cm = metrics.confusion_matrix(y_true, y_pred, labels=classes, normalize=None)
    plot_confusion_matrix(cm, classes, figsize=figsize)


def plot_confusion_matrix(cm, display_labels, normalize=False, cmap="Blues", figsize=(8, 8),
                          xticks_rotation="vertical", colorbar=True, plot_zero=True, xy_label=True):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    Difference from sklearn.metrics.ConfusionMatrixDisplay:
    - zeros: uses alpha and allows not plotting zeros
    - colorbar: allows disabling colorbar
    """
    # with sns.axes_style("white"):
    fig, ax = plt.subplots(figsize=figsize)
    ax.imshow(cm, interpolation='nearest', cmap=cmap)
    if colorbar:
        ax.colorbar()

    fmt = '.2f' if normalize else 'd'
    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        cnt = cm[i, j]
        if cnt == 0 and not plot_zero:
            continue
        ax.text(j, i, format(cnt, fmt), alpha=1 if cnt > 0 else .3, ha="center", va="center",
                color="white" if cnt > thresh else "black")

    n_classes = cm.shape[0]
    ax.set(xticks=np.arange(n_classes), yticks=np.arange(n_classes),
           xticklabels=display_labels, yticklabels=display_labels)
    if xy_label:
        ax.set(ylabel="True label", xlabel="Predicted label")

    ax.set_ylim((n_classes - 0.5, -0.5))
    plt.setp(ax.get_xticklabels(), rotation=xticks_rotation)
    if xticks_rotation == 45:
        plt.setp(ax.get_xticklabels(), ha="right")

    # otherwise x/y labels are sometimes cut off when saving figure
    plt.tight_layout()


def confusion_classes(cm, le):
    n = cm.shape[0]
    cm_nondiag = np.where(np.eye(n), -1, cm)  # hack: set diagonal entries to -1
    true_le, pred_le = np.where(cm_nondiag > 0)

    return pd.DataFrame({
        'support': [cm[i, j] for i, j in zip(true_le, pred_le)],
        "pred_le": pred_le,
        "true_le": true_le,
        "pred_label": le.inverse_transform(pred_le),
        "true_label": le.inverse_transform(true_le)
    }).sort_values("support", ascending=False)


def test_cm():
    y_true = [2, 0, 2, 2, 0, 1]
    y_pred = [0, 0, 2, 2, 0, 2]
    cm_test = metrics.confusion_matrix(y_true, y_pred)
    plot_confusion_matrix(cm_test, ["a", "b", "c"], figsize=(3, 3))
    le_test = LabelEncoder()
    le_test.fit_transform(["a", "b", "c"])
    confusion_classes(cm_test, le_test)



"""Bounding Box methods"""
import functools

import numpy as np


def _check_bb_shape(bb_tlbr: np.ndarray):
    if len(bb_tlbr) == 0:
        bb_tlbr = np.reshape(bb_tlbr, (0, 4))
    assert bb_tlbr.shape[1] == 4, bb_tlbr.shape
    return bb_tlbr


def _get_pairwise_inter_area(bb_tlbr1: np.ndarray, bb_tlbr2: np.ndarray):
    """Pairwise vectorized bounding box intersection area computation for two arrays with t,l,b,r convention
    Args:
        bb_tlbr1: shape (m,4)
        bb_tlbr2: shape (n,4)
    Returns:
        intersection areas with shape (m,n), bb1 areas (m,), bb2 areas (n,)
    """
    bb_tlbr1 = _check_bb_shape(bb_tlbr1)
    bb_tlbr2 = _check_bb_shape(bb_tlbr2)
    t1, l1, b1, r1 = np.split(bb_tlbr1, 4, axis=1)
    t2, l2, b2, r2 = np.split(bb_tlbr2, 4, axis=1)

    t_max = np.maximum(t1, np.transpose(t2))
    l_max = np.maximum(l1, np.transpose(l2))
    b_min = np.minimum(b1, np.transpose(b2))
    r_min = np.minimum(r1, np.transpose(r2))

    inter_area = np.maximum((b_min - t_max), 0) * np.maximum((r_min - l_max), 0)
    bb1_area = (b1 - t1) * (r1 - l1)
    bb2_area = (b2 - t2) * (r2 - l2)
    return inter_area, bb1_area, bb2_area


# https://medium.com/@venuktan/vectorized-intersection-over-union-iou-in-numpy-and-tensor-flow-4fa16231b63d
def iou_matrix(bb_tlbr1: np.ndarray, bb_tlbr2: np.ndarray):
    """Pairwise vectorized iou matrix calculation for two arrays of bounding boxes with t,l,b,r convention
    Args:
        bb_tlbr1: shape (m,4)
        bb_tlbr2: shape (n,4)
    Returns:
        iou matrix with shape (m,n)
    """
    inter_area, bb1_area, bb2_area = _get_pairwise_inter_area(bb_tlbr1, bb_tlbr2)
    iou = inter_area / (bb1_area + np.transpose(bb2_area) - inter_area)
    return iou


def pairwise_intersection_over_area(bb_tlbr1: np.ndarray, bb_tlbr2: np.ndarray):
    """Pairwise intersection over bb1 area calculation for two arrays of bounding boxes with t,l,b,r convention
    Args:
        bb_tlbr1: shape (m,4)
        bb_tlbr2: shape (n,4)
    Returns:
        intersection over area of bb1 with shape (m,n)
    """
    inter_area, bb1_area, bb2_area = _get_pairwise_inter_area(bb_tlbr1, bb_tlbr2)
    return inter_area / bb1_area


def do_matching(match_matrix: np.ndarray, thresh: float):
    match_max = match_matrix.max(axis=1)
    bb2_idxs = match_matrix.argmax(axis=1)
    # set matched idx to -1 if below thresh
    bb2_idxs[match_max < thresh] = -1

    bb1_idxs = np.arange(match_matrix.shape[0])
    return bb1_idxs, bb2_idxs, match_max


def ioa_matching(bb_tlbr1: np.ndarray, bb_tlbr2: np.ndarray, ioa_thresh: float):
    ioa_matrix = pairwise_intersection_over_area(bb_tlbr1, bb_tlbr2)
    return do_matching(ioa_matrix, ioa_thresh)


def iou_matching(bb_tlbr1: np.ndarray, bb_tlbr2: np.ndarray, iou_thresh: float):
    return do_matching(iou_matrix(bb_tlbr1, bb_tlbr2), iou_thresh)


def iou_vector(bb_tlbr1: np.ndarray, bb_tlbr2: np.ndarray):
    """computes iou between the bounding boxes at the same index in bboxes1 and bboxes2
    :param bboxes1 shape (n,4)
    :param bboxes2 shape (n,4)
    :return iou vector with shape (n,)
    """
    assert bb_tlbr1.shape[0] == bb_tlbr2.shape[0]
    ious = [iou_matrix(b1[np.newaxis, :], b2[np.newaxis, :]).squeeze() for b1, b2 in zip(bb_tlbr1, bb_tlbr2)]
    return np.array(ious)


# inspired by:
# https://gamedev.stackexchange.com/questions/154036/efficient-minimum-distance-between-two-axis-aligned-squares
def bbs_distances(bboxes1: np.ndarray, bboxes2: np.ndarray):
    """Vectorized axis-aligned bounding box distance calculation between two arrays of t,l,b,r bounding boxes"""
    t1, l1, b1, r1 = np.split(bboxes1, 4, axis=1)
    t2, l2, b2, r2 = np.split(bboxes2, 4, axis=1)

    t_min = np.minimum(t1, t2.T)
    l_min = np.minimum(l1, l2.T)
    b_max = np.maximum(b1, b2.T)
    r_max = np.maximum(r1, r2.T)

    bb_outer_widths = r_max - l_min
    bb_outer_heights = b_max - t_min

    bbs1_widths = r1 - l1
    bbs1_heights = b1 - t1

    bbs2_widths = r2 - l2
    bbs2_heights = b2 - t2

    inner_width = np.maximum(0, bb_outer_widths - bbs1_widths - bbs2_widths.T)
    inner_height = np.maximum(0, bb_outer_heights - bbs1_heights - bbs2_heights.T)
    min_distance = np.sqrt(inner_width ** 2 + inner_height ** 2)
    return min_distance


def pts_boxes_distance(pts, boxes_ltrb, zero_dist_pt_within_box: bool = True):
    """Distance of a point to a bounding box rectangle
    Based on Stackoverflow: https://stackoverflow.com/a/18157551/1501100
    NOTE: This is an exact solution that leverages the fact that the bounding boxes are axis-aligned
    :param zero_dist_pt_within_box: return zero distance for point in a box (True) or min. distance to all sides (False)
    """
    assert pts.ndim == 2
    assert pts.shape[1] == 2

    # this is pairwise, i.e. measure distance of all points to all boxes
    xs, ys = [p.view(-1, 1) for p in pts.T]
    x_min, y_min, x_max, y_max = [p.view(1, -1) for p in boxes_ltrb.T]

    xmin_d = x_min - xs
    xmax_d = xs - x_max
    ymin_d = y_min - ys
    ymax_d = ys - y_max

    import torch
    dx = torch.clamp(torch.max(xmin_d, xmax_d), 0)
    dy = torch.clamp(torch.max(ymin_d, ymax_d), 0)
    ds = torch.sqrt(dx ** 2 + dy ** 2)

    pts_in_box_mask = (ds == 0)
    if not zero_dist_pt_within_box and pts_in_box_mask.any():
        d_min = functools.reduce(torch.min, [xmin_d.abs(), xmax_d.abs(), ymin_d.abs(), ymax_d.abs()])
        ds[pts_in_box_mask] = d_min[pts_in_box_mask]

    return ds


def pairwise_box_in_box(bbs_tlbr1: np.ndarray, bbs_tlbr2: np.ndarray, pad: float = 0):
    """
    Given two lists of boxes B1 and B2 with N and M boxes respectively,
    computes if box B1_i is contained in padded box B2_j between __all__ N x M pairs of boxes.
    NOTE: works for both tlbr and ltrb bounding box convention and is also PyTorch compatible
    """
    assert bbs_tlbr1.ndim == 2 and bbs_tlbr1.shape[1] == 4, bbs_tlbr1.shape
    assert bbs_tlbr2.ndim == 2 and bbs_tlbr2.shape[1] == 4, bbs_tlbr2.shape
    # torch compatibility
    bbs1 = bbs_tlbr1.unsqueeze(1) if hasattr(bbs_tlbr1, "unsqueeze") else bbs_tlbr1[:, np.newaxis, ...]
    bbs2 = bbs_tlbr2.unsqueeze(0) if hasattr(bbs_tlbr2, "unsqueeze") else bbs_tlbr2[np.newaxis, ...]
    isin_min = (bbs1[..., [0, 1]] >= (bbs2[..., [0, 1]] - pad)).all(-1)  # torch/numpy <-> dim/axis
    isin_max = (bbs1[..., [2, 3]] <= (bbs2[..., [2, 3]] + pad)).all(-1)
    b1_in_b2 = isin_min & isin_max
    assert b1_in_b2.shape == (len(bbs_tlbr1), len(bbs_tlbr2))
    return b1_in_b2


def union_boxes(boxes_ltrb):
    import torch
    lt = torch.min(boxes_ltrb[:, [0, 1]], dim=0)[0]
    rb = torch.max(boxes_ltrb[:, [2, 3]], dim=0)[0]
    return torch.cat([lt, rb])


def polygon_box_degree(box: np.ndarray):
    """
    Computes the degree of a bounding box described as a polygon with 4 points.
    Args:
        box: bounding box described through 4 corner points starting from left-top and going clockwise
    Returns: the degree of the rotated bounding box
    """
    assert box.shape == (4, 2), box.shape

    # for horizontal boxes mid1->mid2 is the line from left-mid to right-mid
    mid1 = (box[0] + box[3]) / 2.
    mid2 = (box[1] + box[2]) / 2.

    # see https://stackoverflow.com/questions/41855261/calculate-the-angle-between-a-line-and-x-axis
    # compute angle of mid1->mid2 to x-axis
    xd, yd = mid2 - mid1
    angle = np.arctan2(yd, xd)
    return -np.rad2deg(angle)


def direction_from_degree(degree):
    if -45 < degree < 45:
        return "lr"
    elif 45 <= degree < 135:
        return "bt"
    elif degree <= -135 or degree >= 135:
        return "rl"
    elif -135 < degree <= -45:
        return "tb"
    else:
        raise ValueError(f"Unexpected degree: {degree:.2f}")



from pathlib import Path
from typing import List, Iterator, Union

EXCLUDED_FILE_NAMES = {"__MACOSX", "desktop.ini"}


def glob(path: Path, pattern: str, include_hidden=False) -> List[Path]:
    return _get_sorted_paths(path.glob(pattern), include_hidden)


def ls(path: Union[Path, str], include_hidden=False) -> List[Path]:
    if isinstance(path, str):
        path = Path(path)
    return _get_sorted_paths(path.iterdir(), include_hidden)


def _get_sorted_paths(path_iterator: Iterator[Path], include_hidden: bool) -> List[Path]:
    def filter_fn(p: Path):
        if p.name in EXCLUDED_FILE_NAMES:
            return False
        if include_hidden:
            return True
        if p.name.startswith("."):
            return False
        return not any(parent.name.startswith(".") for parent in p.parents)

    return sorted(filter(filter_fn, path_iterator))



import logging
import os
import random
import warnings
from datetime import datetime
from typing import Union, Collection

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn
from torch.utils.data import DataLoader


def indices_to_mask(indices: torch.Tensor, mask_length: int) -> torch.BoolTensor:
    assert len(indices) == 0 or mask_length > indices.max(), f"mask_length={mask_length} < max(indices)={indices.max()}"
    mask = torch.full((mask_length,), False, dtype=torch.bool, device=indices.device)
    mask[indices] = True
    # noinspection PyTypeChecker
    return mask


# https://github.com/pytorch/pytorch/issues/3025#issuecomment-392601780
# https://github.com/pytorch/pytorch/pull/26144
def isin(element: torch.Tensor, test_elements: Union[Collection, np.ndarray, torch.Tensor]):
    """see numpy isin: https://docs.scipy.org/doc/numpy/reference/generated/numpy.isin.html#numpy.isin
    """
    if not isinstance(test_elements, torch.Tensor):
        if not isinstance(test_elements, list):
            test_elements = list(test_elements)
        test_elements = torch.tensor(test_elements, dtype=element.dtype, device=element.device)
    return (element[..., None] == test_elements).any(-1)


def count_parameters(model: nn.Module):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def score(model: nn.Module, dl: DataLoader, with_loss=True):
    model.eval()

    with torch.no_grad():
        y_scores, y_true = list(zip(*((model(xb), yb) for xb, yb in dl)))
        y_scores, y_true = torch.cat(y_scores), torch.cat(y_true)
        y_pred = torch.argmax(y_scores, dim=1, keepdim=False)

        if with_loss:
            losses = F.cross_entropy(y_scores, y_true, reduction='none')
            return y_true, y_pred, losses
        else:
            return y_true, y_pred


def evaluate(model: nn.Module, valid_dl: DataLoader, split: str):
    model.eval()

    with torch.no_grad():
        y_scores, y_true = [], []
        for xb, yb in valid_dl:
            y_true.append(yb)
            y_scores.append(model(xb))
        return calc_metrics(y_true, y_scores, split)


def calc_metrics(y_true, y_scores, split: str):
    with torch.no_grad():
        # concatenate tensors in case argument is a list of batches
        if isinstance(y_true, list):
            y_true = torch.cat(y_true)
        if isinstance(y_scores, list):
            y_scores = torch.cat(y_scores)

        n_classes = 1 if y_scores.dim == 1 else y_scores.shape[1]

        loss = F.cross_entropy(y_scores, y_true).item()

        y_true = y_true.numpy()
        if n_classes > 1:
            y_pred = torch.argmax(y_scores, dim=1).numpy()
        else:
            y_pred = (y_scores > .5).numpy()

    import sklearn.metrics
    acc = sklearn.metrics.accuracy_score(y_true, y_pred)

    average = "binary" if n_classes == 1 else "macro"
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        p = sklearn.metrics.precision_score(y_true, y_pred, average=average)
        r = sklearn.metrics.recall_score(y_true, y_pred, average=average)

    m = {f"{split}_loss": loss,
         f"{split}_acc": acc,
         f"{split}_prec": p,
         f"{split}_rec": r}
    return m


# copied from detectron2
# https://github.com/facebookresearch/detectron2/blob/master/detectron2/utils/env.py
def seed_all_rng(seed=None):
    """
    Set the random seed for the RNG in torch, numpy and python.

    Args:
        seed (int): if None, will use a strong random seed.
    """
    if seed is None:
        seed = os.getpid() + int(datetime.now().strftime("%S%f")) + int.from_bytes(os.urandom(2), "big")
        logger = logging.getLogger(__name__)
        logger.info("Using a generated random seed {}".format(seed))
    np.random.seed(seed)
    torch.set_rng_state(torch.manual_seed(seed).get_state())
    random.seed(seed)



import copy
import functools
import logging
import math
from collections import Counter
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Union, Dict, Any

import matplotlib as mpl
import matplotlib.cm as cmx
import matplotlib.colors as colors
import matplotlib.patches as mpatches
import matplotlib.path as mpath
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
from matplotlib import patheffects

_logger = logging.getLogger(__name__)

DEFAULT_FIGSIZE = (12, 12)


class BoundingBox:
    """
    This Bounding Box implementation assumes the coordinates in between pixels.
    This means that e.g. the width of the bbox is right-left.
    So a bounding box that contains the top-left pixel would be BoundingBox(0,0,1,1).
    This is in contrast to the convention where BoundingBox(0,0,0,0) corresponds to first pixel, and w = r-l+1
    For some background see: https://github.com/facebookresearch/Detectron/blob/master/detectron/utils/boxes.py#L23
    """

    def __init__(self, t: float, l: float, b: float, r: float, allow_neg_coord=False):
        self.t, self.l, self.b, self.r = t, l, b, r
        self.allow_neg_coord = allow_neg_coord

        assert t <= b, f"Invalid bounding box coordinates: {self}"
        assert l <= r, f"Invalid bounding box coordinates: {self}"

        if any(x < 0 for x in [t, l, b, r]):
            msg = f"Bounding box has coordinates <= 0: {self}"
            if allow_neg_coord:
                _logger.debug(msg)
            else:
                raise ValueError(msg)

    def __eq__(self, other):
        if not isinstance(other, BoundingBox):
            return False
        return self.tlbr == other.tlbr

    def __hash__(self):
        return hash(self.tlbr)

    def __repr__(self):
        return f"BoundingBox(t={self.t:.2f},l={self.l:.2f},b={self.b:.2f},r={self.r:.2f})"

    @classmethod
    def clipped_to_image(cls, t: float, l: float, b: float, r: float, img_w: int, img_h: int):
        return BoundingBox(t=max(t, 0), l=max(l, 0), b=min(b, img_h), r=min(r, img_w))

    @classmethod
    def from_center_wh(cls, center, width, height, allow_neg_coord=False):
        x, y = center
        return cls(
            t=y - height / 2,
            l=x - width / 2,
            b=y + height / 2,
            r=x + width / 2,
            allow_neg_coord=allow_neg_coord
        )

    @classmethod
    def from_xywh(cls, x, y, width, height, allow_neg_coord=False):
        return cls(t=y, l=x, b=y + height, r=x + width, allow_neg_coord=allow_neg_coord)

    @classmethod
    def from_ltrb(cls, ltrb, allow_neg_coord=False):
        l, t, r, b = ltrb
        return cls(t, l, b, r, allow_neg_coord)

    @classmethod
    def from_pascal_voc(cls, xmin, ymin, xmax, ymax, allow_neg_coord=False):
        # pascal has pixel-based boxes in range [1, W/H]
        # convert to continuous coordinate boxes in range [0,W/H]
        return cls(t=ymin - 1, l=xmin - 1, b=ymax, r=xmax, allow_neg_coord=allow_neg_coord)

    @classmethod
    def from_points(cls, pts: np.ndarray, allow_neg_coord=False):
        """
        :param pts: xy pts array assuming top-left origin
        :param allow_neg_coord: allow negative coordinates outside of image
        """
        assert pts.ndim == 2
        assert pts.shape[1] == 2
        l, t = pts.min(axis=0)
        r, b = pts.max(axis=0) + 1  # +1 to convert from pixel to coordinate-based representation
        return BoundingBox(t, l, b, r, allow_neg_coord)

    @classmethod
    def from_mask(cls, mask: np.ndarray, assert_any=False):
        """
        :param mask: 2-d binary mask
        :param assert_any: assert that at least one mask entry is True
        """
        assert mask.ndim == 2
        ys, xs = np.where(mask)
        if assert_any:
            assert len(ys) > 0, "mask has no positive entries"
        pts = np.stack([xs, ys], axis=1)
        return BoundingBox.from_points(pts), (ys, xs)

    @property
    def tlbr(self):
        return self.t, self.l, self.b, self.r

    @property
    def w(self):
        return self.r - self.l

    @property
    def h(self):
        return self.b - self.t

    @property
    def area(self):
        return self.w * self.h

    @property
    def diameter(self):
        return math.sqrt(self.w ** 2 + self.h ** 2)

    @property
    def aspect_ratio(self):
        return self.w / self.h

    @property
    def tb_mid(self):
        return self.t + self.h / 2

    @property
    def lr_mid(self):
        return self.l + self.w / 2

    @property
    def center(self):
        return self.lr_mid, self.tb_mid

    @property
    def xy_w_h(self):
        return (self.l, self.t), self.w, self.h

    @property
    def bb_coco(self):
        return self.l, self.t, self.w, self.h

    @property
    def bb_pascal_voc(self):
        """
        :return: bounding box in pascal voc format ltrb, which corresponds to standard bottom-left origin
        """
        return self.l, self.t, self.r, self.b

    @property
    def ltrb(self):
        return self.bb_pascal_voc

    def is_within_img(self, img_w, img_h):
        return self.is_within_bb(BoundingBox(0, 0, b=img_h, r=img_w))

    def is_within_bb(self, bb):
        t, l, b, r = self.tlbr
        return t >= bb.t and l >= bb.l and b <= bb.b and r <= bb.r

    def iou(self, bb) -> float:
        intersection_area = self.intersection(bb).area
        union_area = self.area + bb.area - intersection_area
        return intersection_area / union_area

    def distance(self, bb) -> float:
        bbs1 = np.array(self.tlbr).reshape(1, -1)
        bbs2 = np.array(bb.tlbr).reshape(1, -1)
        from yamlu.bb import bbs_distances
        return bbs_distances(bbs1, bbs2).item()

    def intersection(self, bb):
        t = max(self.t, bb.t)
        l = max(self.l, bb.l)
        b = min(self.b, bb.b)
        r = min(self.r, bb.r)
        if b < t:
            b = t
        if r < l:
            r = l
        return BoundingBox(t=t, l=l, b=b, r=r, allow_neg_coord=self.allow_neg_coord)

    def union(self, bb):
        return BoundingBox(t=min(self.t, bb.t), l=min(self.l, bb.l), b=max(self.b, bb.b), r=max(self.r, bb.r),
                           allow_neg_coord=self.allow_neg_coord)

    def pad(self, pad, allow_neg_coord=False) -> "BoundingBox":
        return BoundingBox(self.t - pad, self.l - pad, self.b + pad, self.r + pad, allow_neg_coord=allow_neg_coord)

    def pad_min_size(self, w_min, h_min):
        w_new = max(self.w, w_min)
        h_new = max(self.h, h_min)
        return BoundingBox.from_center_wh(self.center, w_new, h_new)

    def shrink(self, px) -> "BoundingBox":
        return self.pad(-px)

    def shift(self, t_delta=0, l_delta=0) -> "BoundingBox":
        t, l, b, r = self.tlbr
        return BoundingBox(t=t + t_delta, l=l + l_delta, b=b + t_delta, r=r + l_delta,
                           allow_neg_coord=self.allow_neg_coord)

    def clip_to_image(self, img_w: int, img_h: int) -> "BoundingBox":
        t, l, b, r = self.tlbr
        return BoundingBox(t=max(t, 0), l=max(l, 0), b=min(b, img_h), r=min(r, img_w))

    def scale(self, factor) -> "BoundingBox":
        tlbr = np.array(self.tlbr) * factor
        return BoundingBox(*tlbr, allow_neg_coord=self.allow_neg_coord)

    def rotate(self, angle, img_size) -> "BoundingBox":
        assert angle % 90 == 0 and angle >= 0, f"Invalid angle: {angle}"
        img_w, img_h = img_size
        if angle == 0:
            return self
        else:
            bb = self.rot90(img_h)
            return bb.rotate(angle - 90, (img_h, img_w))

    def rot90(self, img_h) -> "BoundingBox":
        return BoundingBox(
            t=self.l,
            l=img_h - self.b,
            b=self.r,
            r=img_h - self.t
        )


class Annotation:
    def __init__(self, category: str, bb: BoundingBox, **kwargs: Any):
        self._fields: Dict[str, Any] = {
            "category": category,
            "bb": bb,
            **kwargs
        }

    @property
    def category(self) -> str:
        return self._fields["category"]

    @property
    def bb(self) -> BoundingBox:
        return self._fields["bb"]

    @property
    def extra_fields(self):
        return {k: v for k, v in self._fields.items() if k not in ["category", "bb"]}

    def plot(self, img: Image.Image, figsize=None, pad=30, **plot_ann_kwargs):
        plot_img(img, figsize=figsize)
        plot_anns([self], **plot_ann_kwargs)

        bb = self.bb
        plt.xlim(bb.l - pad, bb.r + pad)
        plt.ylim(bb.b + pad, bb.t - pad)

    def img_cropped(self, img: Union[Image.Image, np.ndarray], pad=0, return_bb=False):
        img = np.asarray(img)
        t, l, b, r = self.bb.tlbr
        t, l = [max(math.floor(x) - pad, 0) for x in [t, l]]

        img_h, img_w = img.shape[:2]
        b, r = [min(math.ceil(x) + pad, m - 1) for x, m in zip([b, r], [img_h, img_w])]

        img_crop = Image.fromarray(img[t:b, l:r, ...])
        if return_bb:
            return img_crop, BoundingBox(t, l, b, r)
        return img_crop

    def __setattr__(self, name: str, val: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, val)
        else:
            self._fields[name] = val

    def __getattr__(self, name: str) -> Any:
        if name.startswith("_"):
            return super().__getattribute__(name)
        return self._fields[name]

    def __delattr__(self, name: str):
        if name.startswith("_"):
            return super().__delattr__(name)
        del self._fields[name]

    def set(self, name: str, value: Any):
        setattr(self, name, value)

    def get(self, name: str):
        return self._fields[name]

    def __contains__(self, key):
        return key in self._fields

    def __repr__(self):
        # only get fields that are not of type Annotation to prevent potentially infinite recursion
        fields_str = ", ".join(f"{k}={v}" for k, v in self.extra_fields.items() if not isinstance(v, Annotation))
        return f"{self.__class__.__name__}(category='{self.category}', bb={self.bb}, {fields_str})"


@dataclass
class AnnotatedImage:
    filename: str
    width: int
    height: int
    annotations: List[Annotation]
    img: Optional[Image.Image] = field(default=None, repr=False)

    @classmethod
    def from_img_path(cls, img_path: Path, annotations: List[Annotation]):
        img = read_img(img_path)
        return cls(img_path.name, width=img.width, height=img.height, annotations=annotations, img=img)

    def __len__(self):
        return len(self.annotations)

    def __getitem__(self, idx):
        return self.annotations[idx]

    def copy(self) -> "AnnotatedImage":
        """
        Creates a copy of the AnnotatedImage, but without copying the image itself
        """
        return AnnotatedImage(
            filename=self.filename,
            width=self.width,
            height=self.height,
            annotations=copy.deepcopy(self.annotations),
            img=self.img
        )

    def plot(self, figsize=DEFAULT_FIGSIZE, **plot_anns_kwargs):
        assert self.img is not None, f"{self}: missing img attribute!"
        plot_img(self.img, figsize=figsize)
        plot_anns(self.annotations, **plot_anns_kwargs)

    def save(self, directory: Path, **params):
        directory.mkdir(exist_ok=True, parents=True)
        self.img.save(directory / self.filename, **params)

    def save_with_anns(self, directory: Path, figsize=None, suffix="_bb", jpg_quality=75, **plot_anns_kwargs):
        self.plot(figsize=figsize, **plot_anns_kwargs)
        directory.mkdir(exist_ok=True, parents=True)
        img_path = directory / f"{self.img_id}{suffix}.jpg"
        plt.savefig(str(img_path), pil_kwargs={"quality": jpg_quality})
        plt.close()
        return img_path

    def filter(self, category: str, *other_categories) -> List[Annotation]:
        assert isinstance(category, str), f"Wrong type for category {category}: {type(category)}"
        return [a for a in self.annotations if a.category in {category, *other_categories}]

    def filter_substr(self, substr: str):
        return [a for a in self.annotations if substr in a.category]

    def union_bb(self):
        bbs = [a.bb for a in self.annotations]
        return functools.reduce(lambda bb1, bb2: bb1.union(bb2), bbs)

    @property
    def boxes_tlbr(self):
        boxes = np.array([a.bb.tlbr for a in self.annotations])
        # reshape so that shape is (0,4) for images with no annotations
        return boxes.reshape(len(self.annotations), 4)

    @property
    def boxes_ltrb(self):
        return self.boxes_tlbr[:, [1, 0, 3, 2]]

    @property
    def size(self):
        return self.width, self.height

    @property
    def img_id(self) -> str:
        return Path(self.filename).stem

    @property
    def categories(self):
        return [a.category for a in self.annotations]

    def __repr__(self):
        n_anns = len(self.annotations)
        cat_cnt = Counter(a.category for a in self.annotations)
        s = f"AnnotatedImage(filename='{self.filename}', size={self.size}, {n_anns} annotations ({cat_cnt}))"
        return s


def compute_colors(annotations: List[Annotation], categories: List[str], cmap='jet') -> List[np.ndarray]:
    """
    :param annotations: list of annotations
    :param categories: dataset categories
    :param cmap: matplotlib cmap (other options include Dark2, Accent)
    :return: a list with an rgba array for each annotation
    """
    dataset_cat_to_id = {c: i for i, c in enumerate(categories)}

    cat_ids = np.array([dataset_cat_to_id[ann.category] for ann in annotations])

    cm = plt.get_cmap(cmap)
    cNorm = colors.Normalize(vmin=0, vmax=len(categories) - 1)
    scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=cm)
    scalarMap.get_clim()

    return [scalarMap.to_rgba(c) for c in cat_ids]


def plot_anns(annotations: List[Annotation], categories: List[str] = None, ann_colors=None, ax=None, with_index=False,
              with_score=True, digits: int = 1, min_score: float = 0.0, draw_connections=True, draw_keypoints=True,
              alpha_bb=.15, alpha_txt=.5, alpha_kp=.3,
              font_size_scale=.7, lw_scale=.1, black_lw_font_scale=0.05, text_horizontal_pos="left",
              text_vertical_pos="top", text_field="category"):
    """
    :param annotations: annotations to plot
    :param categories: dataset categories for computing deterministic annotation colors
    :param ann_colors: color for each annotation to use (categories is ignored if ann_colors is set) or single color
    :param ax: matplotlib axes
    :param with_index: include the annotation index into the bounding box label
    :param with_score: include the score into the bounding box label
    :param digits: score digits
    :param min_score: score threshold for plotting the annotation
    :param draw_connections: draw arrow connections
    :param alpha_bb: alpha for the bounding box fill (not the border)
    :param alpha_txt: alpha for the text
    :param alpha_kp: alpha for scattered keypoints
    :param font_size_scale: font size scale relative to larger size of the figure
    :param lw_scale: bounding box linewidth relative to larger size of the figure
    :param black_lw_font_scale: scale of the black stroke surrounding the font relative to font size
    :param text_horizontal_pos: horizontal position of bounding box text (one of 'left', 'center', 'right')
    :param text_vertical_pos: vetical position of bounding box text (one of 'top', 'bottom')
    :param text_field: field to use for the bounding box title
    """
    if len(annotations) == 0:
        _logger.warning("plot_anns: passed empty annotations list")
        return

    if ax is None:
        ax = plt.gca()
    if categories is None:
        categories = list(set(a.category for a in annotations))
    if ann_colors is None:
        ann_colors = compute_colors(annotations, categories)
    elif not isinstance(ann_colors, list):
        # assume only a single color has been passed
        ann_colors = [ann_colors] * len(annotations)

    annotations = [a for a in annotations if "score" not in a or a.score >= min_score]

    # rough estimates so that text + kp sizes scale with figure size
    figsize = ax.figure.get_size_inches()
    larger_size = max(figsize)
    fontsize = max(larger_size * font_size_scale, 8)
    conn_size = larger_size * .3
    lw = larger_size * lw_scale
    kp_size = (fontsize * 0.66) ** 2  # in plt.scatter s is area (w*h)
    black_lw_font = fontsize * black_lw_font_scale

    for i, ann, color in zip(range(len(annotations)), annotations, ann_colors):
        patch = mpatches.Rectangle(*ann.bb.xy_w_h, fill=True, facecolor=color, edgecolor=color, lw=0, alpha=alpha_bb)
        ax.add_patch(patch)
        patch = mpatches.Rectangle(*ann.bb.xy_w_h, fill=False, facecolor="none", edgecolor=color, lw=lw, alpha=.8)
        ax.add_patch(patch)

        text = "" if text_field is None or text_field not in ann else getattr(ann, text_field)
        if with_score and "score" in ann:
            text += f" {round(ann.score * 100, digits)}%".replace(".0", "")
        if with_index:
            text += f" {i}"

        txt_kwargs = _txt_coord_alignment(ann.bb, text_horizontal_pos, text_vertical_pos)
        txt = ax.text(**txt_kwargs, s=text, color=color, fontsize=fontsize, alpha=alpha_txt)
        if black_lw_font > 0:
            txt.set_path_effects(
                [patheffects.Stroke(linewidth=black_lw_font, foreground='BLACK'), patheffects.Normal()])

        if draw_keypoints:
            if "keypoints" in ann or "waypoints" in ann:
                kps = ann.keypoints if "keypoints" in ann else ann.waypoints
                xs, ys = kps.T[:2]
                # ax.plot(xs, ys, ".-", markersize=fontsize * 0.66, alpha=alpha_kp, linewidth=conn_size, color="red")
                ax.scatter(xs, ys, s=kp_size, alpha=alpha_kp, color="red")
            if "head" in ann:
                ax.scatter(*ann.head, marker=">", s=kp_size, alpha=alpha_kp, color="blue", edgecolor="black", lw=1)
            if "tail" in ann:
                ax.scatter(*ann.tail, marker="o", s=kp_size, alpha=alpha_kp, color="blue", edgecolor="black", lw=1)
        if draw_connections:
            draw_arrow_connections(ann, ax, lw_conn=conn_size, head_length=conn_size * 2, head_width=conn_size)


def _txt_coord_alignment(bb: BoundingBox, text_horizontal_pos: str, text_vertical_pos: str):
    assert text_horizontal_pos in ["left", "center", "right"]
    assert text_vertical_pos in ["top", "center", "bottom"]
    if text_horizontal_pos == "left":
        horiz_coord = bb.l
    elif text_horizontal_pos == "center":
        horiz_coord = bb.lr_mid
    else:
        horiz_coord = bb.r

    # text vertical alignment is opposite of positioning, i.e. text is aligned to bottom when it's placed on top of bb
    if text_vertical_pos == "top":
        va = "bottom"
        vert_coord = bb.t
    elif text_vertical_pos == "bottom":
        va = "top"
        vert_coord = bb.b
    elif text_vertical_pos == "center":
        va = "center"
        vert_coord = bb.tb_mid

    # text horizontal alignment equals the position
    ha = text_horizontal_pos

    return {"x": horiz_coord, "y": vert_coord, "va": va, "ha": ha}


def draw_arrow_connections(ann: Annotation, ax, lw_conn=4, color=(220 / 255., 20 / 255., 60 / 255.),
                           ls="-", head_length=10, head_width=4, alpha=.3):
    vertices = []

    if "arrow_prev" in ann:
        vertices.append(ann.arrow_prev.bb.center)

    if "waypoints" in ann:
        vertices += ann.waypoints.tolist()
    elif "keypoints" in ann:
        vertices += ann.keypoints[:, :2].tolist()
    else:
        kps = [kp for kp in ["tail", "head"] if kp in ann]
        for kp in kps:
            vertices.append(ann.get(kp))

    if "arrow_next" in ann:
        vertices.append(ann.arrow_next.bb.center)

    if len(vertices) <= 1:
        return

    codes = [mpath.Path.MOVETO] + [mpath.Path.LINETO] * (len(vertices) - 1)

    path = mpath.Path(vertices, codes)

    arw_style = mpatches.ArrowStyle.CurveFilledB(head_width=head_width, head_length=head_length)
    # noinspection PyTypeChecker
    p = mpatches.FancyArrowPatch(
        path=path,
        arrowstyle=arw_style,
        linewidth=lw_conn,
        linestyle=ls,
        color=color,
        alpha=alpha,
        # path_effects=[patheffects.Stroke(linewidth=2, foreground='BLACK', alpha=alpha), patheffects.Normal()]
    )
    # TODO capstyle and joinstyle https://matplotlib.org/3.1.0/gallery/lines_bars_and_markers/joinstyle.html
    ax.add_patch(p)


def plot_img(img: Union[np.ndarray, Image.Image], figsize=DEFAULT_FIGSIZE, save_path=None, **imshow_kwargs):
    if hasattr(img, "cpu"):  # check if pytorch tensor without importing torch
        img = img.cpu().numpy()

    if not figsize:
        figsize = figsize_from_img(img)

    fig = plt.figure(figsize=figsize)

    ax = fig.add_axes([0, 0, 1, 1])
    ax.axis("off")
    # hard-coded cmap gray: seems to have no effect for RGB images but is required for grayscale images
    ax.imshow(np.asarray(img), cmap="gray", **imshow_kwargs)

    if save_path:
        plt.savefig(save_path)


def plot_imgs(imgs: Union[np.ndarray, List[np.ndarray], List[Image.Image]], ncols: int = None,
              img_size=(5, 5), axis_opt="off", titles: List[str] = None, max_allowed_imgs=100,
              **imshow_kwargs):
    """
    :param imgs: batch of imgs with shape (batch_size, h, w) or (batch_size, h, w, 3)
    :param ncols: number of columns
    :param img_size: matplotlib size to use for each image
    :param axis_opt: plot axis or not ("off"/"on")
    :param titles: axis titles
    :param max_allowed_imgs: throws an error if more than max_allowed_imgs are passed
    """
    n_imgs = len(imgs)
    assert 0 < n_imgs < max_allowed_imgs, f"n_imgs: {n_imgs}"

    if ncols is None:
        ncols = min(n_imgs, 5)

    if titles is not None:
        assert len(imgs) == len(titles), f"{len(imgs)} != {len(titles)}"

    if hasattr(imgs, "detach"):  # check if pytorch tensor without importing torch
        imgs = imgs.detach().cpu().numpy()

    nrows = math.ceil(n_imgs / ncols)
    img_w, img_h = img_size
    figsize = img_w * ncols, img_h * nrows

    nrows = math.ceil(n_imgs / ncols)
    fig, axs = plt.subplots(nrows, ncols, figsize=figsize)
    for i, img in enumerate(imgs):
        if nrows == 1 or ncols == 1:
            ax: plt.Axes = axs[i]
        else:
            ax: plt.Axes = axs[i // ncols, i % ncols]
        ax.axis(axis_opt)
        ax.imshow(np.asarray(img), **imshow_kwargs)
        if titles is not None:
            ax.set_title(titles[i])

    # this is quite slow:
    # fig.tight_layout()
    return axs


def plot_img_paths(img_paths: Union[List[Path], List[str]], ncols=4, img_size=(5, 5)):
    imgs = [read_img(p) for p in img_paths]
    plot_imgs(imgs, ncols=ncols, img_size=img_size, titles=[Path(p).name for p in img_paths])


def read_img(img_path: Union[Path, str]) -> Image.Image:
    img = Image.open(img_path)
    return exif_transpose(img)


# same as ImageOps.exif_transpose, but does not copy the image, which avoids Image.load()
def exif_transpose(image) -> Image.Image:
    """
    If an image has an EXIF Orientation tag, return a new image that is
    transposed accordingly. Otherwise, return the image.

    :param image: The image to transpose.
    :return: An image.
    """
    exif = image.getexif()
    orientation = exif.get(0x0112)
    method = {
        2: Image.FLIP_LEFT_RIGHT,
        3: Image.ROTATE_180,
        4: Image.FLIP_TOP_BOTTOM,
        5: Image.TRANSPOSE,
        6: Image.ROTATE_270,
        7: Image.TRANSVERSE,
        8: Image.ROTATE_90,
    }.get(orientation)
    if method is not None:
        transposed_image = image.transpose(method)
        del exif[0x0112]
        transposed_image.info["exif"] = exif.tobytes()
        return transposed_image
    return image


def figsize_from_img(img: Union[Image.Image, np.ndarray]):
    if isinstance(img, np.ndarray):
        h, w = img.shape[:2]
        return figsize_from_wh(w, h)
    elif isinstance(img, Image.Image):
        return figsize_from_wh(*img.size)
    else:
        raise ValueError(f"Unknown image type: {type(img)}")


def figsize_from_wh(img_w, img_h):
    # What size does the figure need to be in inches to fit the image?
    dpi = mpl.rcParams['figure.dpi']
    return img_w / float(dpi), img_h / float(dpi)



import torch

from yamlu.pytorch import isin, indices_to_mask


# noinspection PyArgumentList
def test_indices_to_mask():
    indices = torch.LongTensor([0, 2, 3])
    mask_length = 5
    mask = indices_to_mask(indices, mask_length)
    assert torch.equal(mask, torch.BoolTensor([True, False, True, True, False]))


# noinspection PyArgumentList
def test_isin():
    element = torch.LongTensor([0, 1, 3, 2, 1, 2])
    test_elements = torch.LongTensor([0, 1])
    res = isin(element, test_elements)
    assert res.tolist() == [1, 1, 0, 0, 1, 0]

    res = isin(element.to(torch.int), test_elements.to(torch.int))
    assert res.tolist() == [1, 1, 0, 0, 1, 0]

    res = isin(element, [0, 1])
    assert res.tolist() == [1, 1, 0, 0, 1, 0]



from yamlu.img import read_img, plot_img, plot_imgs, plot_img_paths, plot_anns
from yamlu.misc import flatten
from yamlu.np_utils import bin_stats
from yamlu.path import ls, glob

all = [
    "read_img", "plot_img", "plot_imgs", "plot_img_paths", "plot_anns",
    "flatten",
    "bin_stats",
    "ls", "glob",
]

try:
    from yamlu.pytorch import isin

    all.append("isin")
except ImportError:
    pass

__all__ = all



import matplotlib.patheffects as PathEffects
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from sklearn.preprocessing import LabelEncoder


def tsne_scatter(X, Y, perplexity=40, n_iter=400):
    """
    inspired by: https://www.datacamp.com/community/tutorials/introduction-t-sne
    """
    tsne = TSNE(n_components=2, verbose=0, perplexity=perplexity, n_iter=n_iter)
    X_tsne = tsne.fit_transform(X)

    num_classes = len(np.unique(Y))

    # choose a color palette with seaborn.
    palette = np.array(sns.color_palette("hls", num_classes))

    le = LabelEncoder()
    y_le = le.fit_transform(Y)

    # create a scatter plot.
    f = plt.figure(figsize=(8, 8))
    ax: plt.Axes = plt.subplot(aspect='equal')
    ax.scatter(X_tsne[:, 0], X_tsne[:, 1], lw=0, s=40, c=palette[y_le.astype(np.int)])
    plt.xlim(-25, 25)
    plt.ylim(-25, 25)
    ax.axis('off')
    ax.axis('tight')

    # add the labels for each digit corresponding to the label
    txts = []

    for i in range(num_classes):
        # Position of each label at median of data points.
        xtext, ytext = np.median(X_tsne[y_le == i, :], axis=0)
        txt = ax.text(xtext, ytext, le.inverse_transform([i])[0], fontsize=24)
        txt.set_path_effects([
            PathEffects.Stroke(linewidth=5, foreground="w"),
            PathEffects.Normal()])
        txts.append(txt)

    return f, ax


def tsne_pca_scatter(X, Y, pca_dims=50):
    pca = PCA(n_components=pca_dims)
    X_pca = pca.fit_transform(X)
    tsne_scatter(X_pca, Y)



from setuptools import setup, find_packages

setup(
    name="yamlu",
    version="0.0.7",
    description="yet another machine learning utility library",
    url="https://github.com/bernhardschaefer/yamlu",
    author="Bernhard Schäfer",
    author_email="bernhard.schaefer1+pypi@gmail.com",
    license="Apache 2.0",
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=[
        "matplotlib",
        "numpy",
        "Pillow"
    ]
)



from typing import Union, Tuple

import numpy as np
from PIL import Image
from matplotlib import colors


def grayscale_transparency(img: Image.Image) -> Image.Image:
    """
    Sets the alpha transparency of an image according to its grayscale value,
    i.e. white is fully transparent whereas black is not transparent.
    """
    # noinspection PyTypeChecker
    img_np = np.asarray(img).copy()
    # noinspection PyTypeChecker
    grayscale = np.asarray(img.convert("L"))
    img_np[:, :, -1] = 255 - grayscale
    return Image.fromarray(img_np)


def white_to_transparency(img: Image.Image, thresh=255) -> Image.Image:
    """
    Makes the white pixels in an image transparent
    :param img: source image
    :param thresh: pixels where all RGB values are higher or equal than this threshold are considered white
    """
    # FIXME change implementation so that it selects white pixels and only modifies their alpha
    # noinspection PyTypeChecker
    x = np.asarray(img.convert('RGBA')).copy()
    # inspired by https://stackoverflow.com/a/54148416
    non_white_mask = (x[:, :, :3] < thresh).any(axis=2)
    x[:, :, 3] = (255 * non_white_mask).astype(np.uint8)
    return Image.fromarray(x)


def black_to_color(img: Image.Image, color: Union[str, Tuple[int, int, int]], thresh=128) -> Image.Image:
    """
    Converts black pixels in an image to another color
    :param img: source image
    :param color: target color to apply to black pixels
    :param thresh: a pixel value threshold, each pixel with intensity less than thresh will be converted
    """
    rgb = (np.array(colors.to_rgb(color)) * 255.).astype(np.uint8)

    # noinspection PyTypeChecker
    x = np.asarray(img).copy()
    black_mask = (x[:, :, :3] <= thresh).all(axis=2)
    x[black_mask, :3] = rgb

    return Image.fromarray(x)



import logging
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Union, Tuple, List, Dict
from xml.dom import minidom

import yamlu
from yamlu.img import AnnotatedImage, BoundingBox, Annotation
from yamlu.np_utils import to_python_type

_logger = logging.getLogger(__name__)


def dump_ai_voc(ai: AnnotatedImage, folder: Path, additional_fields: Tuple = ()) -> Path:
    root = ET.Element("annotation")
    ET.SubElement(root, "folder").text = folder.name
    ET.SubElement(root, "filename").text = ai.filename
    ET.SubElement(root, "path").text = str(folder / ai.filename)

    size = ET.SubElement(root, "size")
    ET.SubElement(size, "width").text = str(ai.width)
    ET.SubElement(size, "height").text = str(ai.height)
    # ET.SubElement(size, "depth").text = ai.height

    for a in ai.annotations:
        obj = ET.SubElement(root, "object")
        ET.SubElement(obj, "name").text = a.category
        ET.SubElement(obj, "difficult").text = str(0)
        ET.SubElement(obj, "occluded").text = str(0)

        for k in additional_fields:
            if k in a:
                v = getattr(a, k)
                if isinstance(v, Annotation):
                    v = v.aid
                ET.SubElement(obj, k).text = str(v)

        bndbox = ET.SubElement(obj, "bndbox")
        bb = a.bb
        # Pascal annotations pixel-based integers in the range [1, W or H],
        # where a box with annotation (xmin=1, xmax=W) covers the whole image.
        # In coordinate space this is represented by (xmin=0, xmax=W)
        for k, v in zip(["xmin", "ymin", "xmax", "ymax"], [bb.l + 1, bb.t + 1, bb.r, bb.b]):
            ET.SubElement(bndbox, k).text = str(to_python_type(v, 0))

    folder.mkdir(parents=True, exist_ok=True)
    xml_path = folder / f"{ai.img_id}.xml"
    # tree = ET.ElementTree(root)
    # tree.write(str(xml_path), encoding="utf-8")

    # without dependencies like lxml I have to reparse wiht minidom to get pretty xml :-/
    # https://stackoverflow.com/questions/28813876/how-do-i-get-pythons-elementtree-to-pretty-print-to-an-xml-file
    xmlstr = minidom.parseString(ET.tostring(root)).toprettyxml(indent="   ")
    xml_path.write_text(xmlstr)
    return xml_path


def parse_voc_annotations(voc_xml_path: Union[str, Path]) -> List[Annotation]:
    tree = ET.parse(voc_xml_path)
    root = tree.getroot()
    # filename = root.find('filename').text

    anns = []
    for obj in root.iter('object'):
        category = obj.find("name").text

        bbox = obj.find("bndbox")
        bb_dict = {e.tag: float(e.text) for e in bbox}
        # we leverage the fact that voc uses xmin, ymin, xmax, ymax field names
        bb = BoundingBox.from_pascal_voc(**bb_dict)

        # Note: difficult and occluded is not parsed since I don't use them at the moment
        add_elements = [e for e in obj if e.tag not in ["name", "bndbox", "difficult", "occluded"]]
        add_fields = {e.tag: e.text for e in add_elements}
        add_fields = {k: int(v) if v.isdigit() else v for k, v in add_fields.items()}

        anns.append(Annotation(category=category, bb=bb, **add_fields))

    return anns


def parse_voc_anns_directory(root: Path) -> Dict[str, List[Annotation]]:
    xml_paths = yamlu.glob(root, "**/*.xml")
    _logger.info("Parsing %d voc xmls", len(xml_paths))
    return {p.stem: parse_voc_annotations(p) for p in xml_paths}


def parse_voc_xml_img(img_path: Path, voc_xml_path: Union[str, Path]) -> AnnotatedImage:
    annotations = parse_voc_annotations(voc_xml_path)
    return AnnotatedImage.from_img_path(img_path, annotations)
